@@ -22,6 +22,8 @@ from future_builtins import map
 
 import os
 import pexpect
+import platform
+import sys
 import time
 import unittest
 
@@ -38,6 +40,24 @@ class CmdError(Exception):
 
 class FunctionalTestCase(DuplicityTestCase):
 
+    _setsid_w = None
+
+    @classmethod
+    def _check_setsid(cls):
+        if cls._setsid_w is not None:
+            return
+        if platform.platform().startswith('Linux'):
+            # setsid behavior differs between distributions.
+            # If setsid supports -w ("wait"), use it.
+            import subprocess
+            try:
+                with open("/dev/null", "w") as sink:
+                    subprocess.check_call(["setsid", "-w", "ls"], stdout=sink, stderr=sink)
+            except subprocess.CalledProcessError:
+                cls._setsid_w = False
+            else:
+                cls._setsid_w = True
+
     def setUp(self):
         super(FunctionalTestCase, self).setUp()
 
@@ -54,6 +74,7 @@ class FunctionalTestCase(DuplicityTestCase):
         if bl:
             backend_inst.delete(backend_inst.list())
         backend_inst.close()
+        self._check_setsid()
 
     def run_duplicity(self, options=[], current_time=None, fail=None,
                       passphrase_input=[]):
@@ -63,7 +84,13 @@ class FunctionalTestCase(DuplicityTestCase):
         # We run under setsid and take input from /dev/null (below) because
         # this way we force a failure if duplicity tries to read from the
         # console unexpectedly (like for gpg password or such).
-        cmd_list = ["setsid", "duplicity"]
+        if platform.platform().startswith('Linux'):
+            cmd_list = ['setsid']
+            if self._setsid_w:
+                cmd_list.extend(["-w"])
+        else:
+            cmd_list = []
+        cmd_list.extend(["duplicity"])
         cmd_list.extend(options)
         cmd_list.extend(["-v0"])
         cmd_list.extend(["--no-print-statistics"])
@@ -78,15 +105,24 @@ class FunctionalTestCase(DuplicityTestCase):
 
         if not passphrase_input:
             cmdline += " < /dev/null"
-        child = pexpect.spawn('/bin/sh', ['-c', cmdline])
+        child = pexpect.spawn('/bin/sh', ['-c', cmdline], timeout=None)
         for passphrase in passphrase_input:
             child.expect('passphrase.*:')
             child.sendline(passphrase)
-        child.wait()
 
+        # if the command fails, we need to clear its output
+        # so it will terminate cleanly.
+        lines = []
+        while child.isalive():
+            lines.append(child.readline())
         return_val = child.exitstatus
-        # output = child.read()
-        # print "Ran duplicity command: ", cmdline, "\n with return_val: ", return_val, "\n and output:\n", output
+
+#         print >>sys.stderr, "duplicity command:", cmdline
+#         print >>sys.stderr, "current dir:", os.getcwd()
+#         print >>sys.stderr, "....output:"
+#         for line in lines:
+#             print >>sys.stderr, line,
+#         print >>sys.stderr, "....return_val:", return_val
 
         if fail:
             self.assertEqual(30, return_val)
